! WHIZARD 2.2.0 alpha Sep 12 2013
! 
! Copyright (C) 1999-2013 by 
!     Wolfgang Kilian <kilian@physik.uni-siegen.de>
!     Thorsten Ohl <ohl@physik.uni-wuerzburg.de>
!     Juergen Reuter <juergen.reuter@desy.de>
!     
!     with contributions by 
!     Christian Speckner <cnspeckn@googlemail.com> 
!     and  Felix Braam, Sebastian Schmidt, Daniel Wiesler 
!
! WHIZARD is free software; you can redistribute it and/or modify it
! under the terms of the GNU General Public License as published by 
! the Free Software Foundation; either version 2, or (at your option)
! any later version.
!
! WHIZARD is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This file has been stripped of most comments.  For documentation, refer
! to the source 'whizard.nw'

module sf_lhapdf

  use kinds, only: default !NODEP!
  use iso_varying_string, string_t => varying_string !NODEP!
  use system_dependencies, only: LHAPDF_PDFSETS_PATH !NODEP!
  use system_dependencies, only: LHAPDF_AVAILABLE !NODEP!
  use limits, only: LHAPDF_DEFAULT_PROTON !NODEP!
  use limits, only: LHAPDF_DEFAULT_PION !NODEP!
  use limits, only: LHAPDF_DEFAULT_PHOTON !NODEP!
  use file_utils !NODEP!
  use diagnostics !NODEP!
  use lorentz !NODEP!
  use unit_tests
  use os_interface
  use models
  use flavors
  use colors
  use quantum_numbers
  use state_matrices
  use polarizations
  use interactions
  use sf_aux
  use sf_base

  implicit none
  private

  public :: lhapdf_status_t
  public :: lhapdf_status_reset
  public :: lhapdf_initialize
  public :: lhapdf_data_t
  public :: sf_lhapdf_test

  type :: lhapdf_status_t
     private
     logical, dimension(3) :: initialized = .false.
  end type lhapdf_status_t

  type, extends (sf_data_t) :: lhapdf_data_t
     private
     type(string_t) :: prefix
     type(string_t) :: file
     integer :: member = 0
     type(model_t), pointer :: model => null ()
     type(flavor_t) :: flv_in
     integer :: set = 0
     logical :: invert = .false.
     logical :: photon = .false.
     logical :: has_photon = .false.
     integer :: photon_scheme = 0
     real(default) :: xmin = 0, xmax = 0
     real(default) :: qmin = 0, qmax = 0
     logical, dimension(-6:6) :: mask = .true.
     logical :: mask_photon = .true.
   contains
       procedure :: init => lhapdf_data_init
       procedure :: write => lhapdf_data_write
       procedure :: get_n_par => lhapdf_data_get_n_par
       procedure :: get_pdg_out => lhapdf_data_get_pdg_out
       procedure :: allocate_sf_int => lhapdf_data_allocate_sf_int
  end type lhapdf_data_t

  type, extends (sf_int_t) :: lhapdf_t
     type(lhapdf_data_t), pointer :: data => null ()
     real(default) :: x = 0
     real(default) :: q = 0
     real(default) :: s = 0
   contains
     procedure :: complete_kinematics => lhapdf_complete_kinematics
     procedure :: inverse_kinematics => lhapdf_inverse_kinematics
     procedure :: type_string => lhapdf_type_string
     procedure :: write => lhapdf_write
     procedure :: init => lhapdf_init
     procedure :: apply => lhapdf_apply
  end type lhapdf_t 
  

  interface
     subroutine InitPDFsetM (set, file)
       integer, intent(in) :: set
       character(*), intent(in) :: file
     end subroutine InitPDFsetM
  end interface

  interface
     subroutine InitPDFM (set, mem)
       integer, intent(in) :: set, mem
     end subroutine InitPDFM
  end interface

  interface
     subroutine numberPDFM (set, n_members)
       integer, intent(in) :: set
       integer, intent(out) :: n_members
     end subroutine numberPDFM
  end interface

  interface
     subroutine evolvePDFM (set, x, q, ff)
       integer, intent(in) :: set
       double precision, intent(in) :: x, q
       double precision, dimension(-6:6), intent(out) :: ff
     end subroutine evolvePDFM
  end interface

  interface
     subroutine evolvePDFphotonM (set, x, q, ff, fphot)
       integer, intent(in) :: set
       double precision, intent(in) :: x, q
       double precision, dimension(-6:6), intent(out) :: ff
       double precision, intent(out) :: fphot
     end subroutine evolvePDFphotonM
  end interface

  interface
     subroutine evolvePDFpM (set, x, q, s, scheme, ff)
       integer, intent(in) :: set
       double precision, intent(in) :: x, q, s
       integer, intent(in) :: scheme
       double precision, dimension(-6:6), intent(out) :: ff
     end subroutine evolvePDFpM
  end interface

  interface
     subroutine GetXminM (set, mem, xmin)
       integer, intent(in) :: set, mem
       double precision, intent(out) :: xmin
     end subroutine GetXminM
  end interface

  interface
     subroutine GetXmaxM (set, mem, xmax)
       integer, intent(in) :: set, mem
       double precision, intent(out) :: xmax
     end subroutine GetXmaxM
  end interface

  interface
     subroutine GetQ2minM (set, mem, q2min)
       integer, intent(in) :: set, mem
       double precision, intent(out) :: q2min
     end subroutine GetQ2minM
  end interface

  interface
     subroutine GetQ2maxM (set, mem, q2max)
       integer, intent(in) :: set, mem
       double precision, intent(out) :: q2max
     end subroutine GetQ2maxM
  end interface

  interface
     function has_photon () result(flag)
        logical :: flag
     end function has_photon
  end interface


contains

  subroutine lhapdf_status_reset (lhapdf_status)
    type(lhapdf_status_t), intent(inout) :: lhapdf_status
    lhapdf_status%initialized = .false.
  end subroutine lhapdf_status_reset

  function lhapdf_status_is_initialized (lhapdf_status, set) result (flag)
    logical :: flag
    type(lhapdf_status_t), intent(in) :: lhapdf_status
    integer, intent(in), optional :: set
    if (present (set)) then
       select case (set)
       case (1:3);    flag = lhapdf_status%initialized(set)
       case default;  flag = .false.
       end select
    else
       flag = any (lhapdf_status%initialized)
    end if
  end function lhapdf_status_is_initialized

  subroutine lhapdf_status_set_initialized (lhapdf_status, set)
    type(lhapdf_status_t), intent(inout) :: lhapdf_status
    integer, intent(in) :: set
    lhapdf_status%initialized(set) = .true.
  end subroutine lhapdf_status_set_initialized

  subroutine lhapdf_initialize (status, set, prefix, file, member)
    type(lhapdf_status_t), intent(inout) :: status
    integer, intent(in) :: set
    type(string_t), intent(inout) :: prefix
    type(string_t), intent(inout) :: file
    integer, intent(inout) :: member
    if (lhapdf_status_is_initialized (status, set))  return
    if (prefix == "")  prefix = LHAPDF_PDFSETS_PATH
    if (file == "") then
       select case (set)
       case (1);  file = LHAPDF_DEFAULT_PROTON
       case (2);  file = LHAPDF_DEFAULT_PION
       case (3);  file = LHAPDF_DEFAULT_PHOTON
       end select
    end if
    if (data_file_exists (prefix // "/" // file)) then
       call InitPDFsetM (set, char (prefix // "/" // file))
    else
       call msg_fatal ("LHAPDF: Data file '" &
            // char (file) // "' not found in '" // char (prefix) // "'.")
       return
    end if
    if (.not. dataset_member_exists (set, member)) then
       call msg_error (" LHAPDF: Chosen member does not exist for set '" &
            // char (file) // "', using default.")
       member = 0
    end if
    call InitPDFM (set, member)
    call lhapdf_status_set_initialized (status, set)
  contains
    function data_file_exists (fq_name) result (exist)
      type(string_t), intent(in) :: fq_name
      logical :: exist
      inquire (file = char(fq_name), exist = exist)
    end function data_file_exists
    function dataset_member_exists (set, member) result (exist)
      integer, intent(in) :: set, member
      logical :: exist
      integer :: n_members
      call numberPDFM (set, n_members)
      exist = member >= 0 .and. member <= n_members
    end function dataset_member_exists
  end subroutine lhapdf_initialize

  subroutine lhapdf_complete_kinematics (sf_int, x, f, r, map)
    class(lhapdf_t), intent(inout) :: sf_int
    real(default), dimension(:), intent(out) :: x
    real(default), intent(out) :: f
    real(default), dimension(:), intent(in) :: r
    logical, intent(in) :: map
    real(default) :: xb1
    if (map) then
       call msg_fatal ("LHAPDF: map flag not supported")
    else
       x(1) = r(1)
       f = 1
    end if
    xb1 = 1 - x(1)
    call sf_int%split_momentum (x, xb1)
    select case (sf_int%status)
    case (SF_DONE_KINEMATICS)
       sf_int%x = x(1)
    case (SF_FAILED_KINEMATICS)
       sf_int%x = 0
       f = 0
    end select
  end subroutine lhapdf_complete_kinematics

  subroutine lhapdf_inverse_kinematics (sf_int, x, f, r, map, set_momenta)
    class(lhapdf_t), intent(inout) :: sf_int
    real(default), dimension(:), intent(in) :: x
    real(default), intent(out) :: f
    real(default), dimension(:), intent(out) :: r
    logical, intent(in) :: map
    logical, intent(in), optional :: set_momenta
    real(default) :: xb1
    logical :: set_mom
    set_mom = .false.;  if (present (set_momenta))  set_mom = set_momenta
    if (map) then
       call msg_fatal ("LHAPDF: map flag not supported")
    else
       r(1) = x(1)
       f = 1
    end if
    xb1 = 1 - x(1)
    if (set_mom) then
       call sf_int%split_momentum (x, xb1)
       select case (sf_int%status)
       case (SF_DONE_KINEMATICS)
          sf_int%x = x(1)
       case (SF_FAILED_KINEMATICS)
          sf_int%x = 0
          f = 0
       end select
    end if
  end subroutine lhapdf_inverse_kinematics

  subroutine lhapdf_data_init &
       (data, status, model, flv, prefix, file, member, photon_scheme)
    class(lhapdf_data_t), intent(out) :: data
    type(lhapdf_status_t), intent(inout) :: status
    type(model_t), intent(in), target :: model
    type(flavor_t), intent(in) :: flv
    type(string_t), intent(in), optional :: prefix, file
    integer, intent(in), optional :: member
    integer, intent(in), optional :: photon_scheme
    integer :: mem
    double precision :: xmin, xmax, q2min, q2max
    external :: InitPDFsetM, InitPDFM, numberPDFM
    external :: GetXminM, GetXmaxM, GetQ2minM, GetQ2maxM
    if (.not. LHAPDF_AVAILABLE) then
       call msg_fatal ("LHAPDF requested but library is not linked")
       return
    end if
    data%model => model
    data%flv_in = flv
    select case (flavor_get_pdg (flv))
    case (PROTON)
       data%set = 1
    case (-PROTON)
       data%set = 1
       data%invert = .true.
    case (PIPLUS)
       data%set = 2
    case (-PIPLUS)
       data%set = 2
       data%invert = .true.
    case (PHOTON)
       data%set = 3
       data%photon = .true.
       if (present (photon_scheme))  data%photon_scheme = photon_scheme
    case default
       call msg_fatal (" LHAPDF: " &
            // "incoming particle must be (anti)proton, pion, or photon.")
       return
    end select
    if (present (prefix)) then
       data%prefix = prefix
    else
       data%prefix = ""
    end if
    if (present (file)) then
       data%file = file
    else
       data%file = ""
    end if
    call lhapdf_initialize &
         (status, data%set, data%prefix, data%file, data%member)
    call GetXminM (data%set, data%member, xmin)
    call GetXmaxM (data%set, data%member, xmax)
    call GetQ2minM (data%set, data%member, q2min)
    call GetQ2maxM (data%set, data%member, q2max)
    data%xmin = xmin
    data%xmax = xmax
    data%qmin = sqrt (q2min)
    data%qmax = sqrt (q2max)
    data%has_photon = has_photon ()
  end subroutine lhapdf_data_init

  subroutine lhapdf_data_write (data, unit) !, md5, beam_fmt)
    class(lhapdf_data_t), intent(in) :: data
    integer, intent(in), optional :: unit
    integer :: u
!    logical, intent(in), optional :: md5
!    logical, intent(in), optional :: beam_fmt
!    logical :: is_md5, is_beam_fmt
!    if (present (md5)) then
!       is_md5 = md5
!    else
!       is_md5 = .false.
!    end if
!    if (present (beam_fmt)) then
!       is_beam_fmt = beam_fmt
!    else
!       is_beam_fmt = .false.
!    end if
    u = output_unit (unit);  if (u < 0)  return
    write (u, "(1x,A)") "LHAPDF data:"
    if (data%set /= 0) then
       write (u, "(3x,A)", advance="no") "flavor       = "
       call flavor_write (data%flv_in, u);  write (u, *)
!       if (.not. is_md5) &
       write (u, "(3x,A,A)")       "  prefix       = ", char (data%prefix)
       write (u, "(3x,A,A)")       "  file         = ", char (data%file)
       write (u, "(3x,A,I3)")      "  member       = ", data%member
       write (u, "(1x,A,ES19.12)") "  x(min)       = ", data%xmin
       write (u, "(1x,A,ES19.12)") "  x(max)       = ", data%xmax
       write (u, "(1x,A,ES19.12)") "  Q^2(min)     = ", data%qmin
       write (u, "(1x,A,ES19.12)") "  Q^2(max)     = ", data%qmax
       write (u, "(3x,A,L1)")      "  invert       = ", data%invert
       if (data%photon)  write (u, "(3x,A,I3)") &
            "  IP2 (scheme) = ", data%photon_scheme
!       if (.not. is_beam_fmt) then          
          write (u, "(3x,A,6(1x,L1),1x,A,1x,L1,1x,A,6(1x,L1))") &
               "  mask         = ", &
               data%mask(-6:-1), "*", data%mask(0), "*", data%mask(1:6)
          write (u, "(3x,A,L1)")   "  photon mask  = ", data%mask_photon
!       end if  
    else
       write (u, "(3x,A)") "[undefined]"       
    end if
  end subroutine lhapdf_data_write
  function lhapdf_data_get_n_par (data) result (n)
    class(lhapdf_data_t), intent(in) :: data
    integer :: n
    n = 1
  end function lhapdf_data_get_n_par
  
  function lhapdf_data_get_pdg_out (data) result (pdg_out)
    class(lhapdf_data_t), intent(in) :: data
    integer, dimension(:), allocatable :: pdg_out
    integer :: n, np, i
    n = count (data%mask)
    np = 0;  if (data%has_photon .and. data%mask_photon)  np = 1
    allocate (pdg_out (n + np))
    pdg_out(1:n) = pack ([(i, i = -6, 6)], data%mask)
    if (np == 1)  pdg_out(n+np) = PHOTON
  end function lhapdf_data_get_pdg_out
  
  subroutine lhapdf_data_allocate_sf_int (data, sf_int)
    class(lhapdf_data_t), intent(in) :: data
    class(sf_int_t), intent(inout), allocatable :: sf_int
    allocate (lhapdf_t :: sf_int)
  end subroutine lhapdf_data_allocate_sf_int
  
  function lhapdf_type_string (object) result (string)
    class(lhapdf_t), intent(in) :: object
    type(string_t) :: string
    if (associated (object%data)) then
       string = "LHAPDF: " // object%data%file
    else
       string = "LHAPDF: [undefined]"
    end if
  end function lhapdf_type_string
  
  subroutine lhapdf_write (object, unit)
    class(lhapdf_t), intent(in) :: object
    integer, intent(in), optional :: unit
    integer :: u
    u = output_unit (unit)
    if (associated (object%data)) then
       call object%data%write (u)
       if (object%status >= SF_DONE_KINEMATICS) then
          write (u, "(1x,A)")  "SF parameters:"
          write (u, "(3x,A,ES17.10)")  "x =", object%x
          if (object%status >= SF_FAILED_EVALUATION) then
             write (u, "(3x,A,ES17.10)")  "Q =", object%q
          end if
       end if
       call object%base_write (u)
    else
       write (u, "(1x,A)")  "LHAPDF data: [undefined]"
    end if
  end subroutine lhapdf_write
    
  subroutine lhapdf_init (sf_int, data)
    class(lhapdf_t), intent(out) :: sf_int
    class(sf_data_t), intent(in), target :: data
    type(quantum_numbers_mask_t), dimension(3) :: mask
    type(flavor_t) :: flv, flv_remnant
    type(quantum_numbers_t), dimension(3) :: qn
    integer :: i
    select type (data)
    type is (lhapdf_data_t)
       mask = new_quantum_numbers_mask (.false., .false., .true.)
       call sf_int%base_init (mask, [0._default], [0._default], [0._default])
       sf_int%data => data
       do i = -6, 6
          if (data%mask(i)) then
             call quantum_numbers_init (qn(1), data%flv_in)
             if (i == 0) then                
                call flavor_init (flv, GLUON, data%model)
                call flavor_init (flv_remnant, HADRON_REMNANT_OCTET, data%model)
             else
                call flavor_init (flv, i, data%model)
                call flavor_init (flv_remnant, &
                     sign (HADRON_REMNANT_TRIPLET, -i), data%model)
             end if
             call quantum_numbers_init (qn(2), &
                  flv = flv_remnant, col = color_from_flavor (flv_remnant, 1))
             call quantum_numbers_init (qn(3), &
                  flv = flv, col = color_from_flavor (flv, 1, reverse=.true.))
             call interaction_add_state (sf_int%interaction_t, qn)
          end if
       end do
       if (data%has_photon .and. data%mask_photon) then
          call flavor_init (flv, PHOTON, data%model)
          call flavor_init (flv_remnant, HADRON_REMNANT_SINGLET, data%model)
          call quantum_numbers_init (qn(2), flv = flv_remnant, &
               col = color_from_flavor (flv_remnant, 1))
          call quantum_numbers_init (qn(3), flv = flv, &
               col = color_from_flavor (flv, 1, reverse=.true.))
          call interaction_add_state (sf_int%interaction_t, qn)
       end if
       call interaction_freeze (sf_int%interaction_t)
       call sf_int%set_incoming ([1])
       call sf_int%set_radiated ([2])
       call sf_int%set_outgoing ([3])
       sf_int%status = SF_INITIAL
    end select
  end subroutine lhapdf_init

  subroutine lhapdf_apply (sf_int, scale)
    class(lhapdf_t), intent(inout) :: sf_int
    real(default), intent(in) :: scale
    real(default) :: x, s
    double precision :: xx, qq, ss    
    double precision, dimension(-6:6) :: ff
    double precision :: fphot
    complex(default), dimension(:), allocatable :: fc
    external :: evolvePDFM, evolvePDFpM
    associate (data => sf_int%data)
      sf_int%q = scale
      x = sf_int%x
      s = sf_int%s
      qq = min (data% qmax, scale)
      qq = max (data% qmin, qq)
      if (.not. data% photon) then
         if (data% invert) then
            if (data%has_photon) then
               call evolvePDFphotonM (data% set, x, qq, ff(6:-6:-1), fphot)
            else
               call evolvePDFM (data% set, x, qq, ff(6:-6:-1))
            end if
         else
            if (data%has_photon) then
               call evolvePDFphotonM (data% set, x, qq, ff, fphot)
            else
               call evolvePDFM (data% set, x, qq, ff)
            end if
         end if
      else
         ss = s
         call evolvePDFpM (data% set, xx, qq, &
              ss, data% photon_scheme, ff)
      end if
      if (data%has_photon) then
         allocate (fc (count ((/data%mask, data%mask_photon/))))
         fc = max (pack ((/ff, fphot/) / x, &
              (/data% mask, data%mask_photon/)), 0._default)
      else
         allocate (fc (count (data%mask)))
         fc = max (pack (ff / x, data%mask), 0._default)
      end if
    end associate
    call interaction_set_matrix_element (sf_int%interaction_t, fc)
    sf_int%status = SF_EVALUATED
  end subroutine lhapdf_apply
  

  subroutine sf_lhapdf_test (u, results)
    integer, intent(in) :: u
    type(test_results_t), intent(inout) :: results
    call test (sf_lhapdf_1, "sf_lhapdf_1", &
         "structure function configuration", &
         u, results)
    call test (sf_lhapdf_2, "sf_lhapdf_2", &
         "structure function instance", &
         u, results)
  end subroutine sf_lhapdf_test
  
  subroutine sf_lhapdf_1 (u)
    integer, intent(in) :: u
    type(os_data_t) :: os_data
    type(model_t), pointer :: model
    type(flavor_t) :: flv
    class(sf_data_t), allocatable :: data
    type(lhapdf_status_t) :: status
    
    write (u, "(A)")  "* Test output: sf_lhapdf_1"
    write (u, "(A)")  "*   Purpose: initialize and display &
         &test structure function data"
    write (u, "(A)")
    
    write (u, "(A)")  "* Create empty data object"
    write (u, "(A)")

    call os_data_init (os_data)
    call syntax_model_file_init ()
    call model_list_read_model (var_str ("QCD"), &
         var_str ("QCD.mdl"), os_data, model)
    call flavor_init (flv, PROTON, model)

    allocate (lhapdf_data_t :: data)
    call data%write (u)

    write (u, "(A)")
    write (u, "(A)")  "* Initialize"
    write (u, "(A)")

    select type (data)
    type is (lhapdf_data_t)
       call data%init (status, model, flv)
    end select

    call data%write (u)

    write (u, "(A)")

    write (u, "(1x,A)")  "Outgoing particle codes:"
    write (u, "(2x,99(1x,I0))")  data%get_pdg_out ()

    
    
    call model_list_final ()
    
    write (u, "(A)")
    write (u, "(A)")  "* Test output end: sf_lhapdf_1"

  end subroutine sf_lhapdf_1

  subroutine sf_lhapdf_2 (u)
    integer, intent(in) :: u
    type(os_data_t) :: os_data
    type(model_t), pointer :: model
    type(flavor_t) :: flv
    class(sf_data_t), allocatable, target :: data
    class(sf_int_t), allocatable :: sf_int
    type(lhapdf_status_t) :: status
    type(vector4_t) :: k
    type(vector4_t), dimension(2) :: q
    real(default) :: E
    real(default), dimension(:), allocatable :: r, x
    real(default) :: f, s
    
    write (u, "(A)")  "* Test output: sf_lhapdf_2"
    write (u, "(A)")  "*   Purpose: initialize and fill &
         &test structure function object"
    write (u, "(A)")
    
    write (u, "(A)")  "* Initialize configuration data"
    write (u, "(A)")

    call os_data_init (os_data)
    call syntax_model_file_init ()
    call model_list_read_model (var_str ("QCD"), &
         var_str ("QCD.mdl"), os_data, model)
    call flavor_init (flv, PROTON, model)

    call reset_interaction_counter ()
    
    allocate (lhapdf_data_t :: data)
    select type (data)
    type is (lhapdf_data_t)
       call data%init (status, model, flv)
    end select
       
    write (u, "(A)")  "* Initialize structure-function object"
    write (u, "(A)")
    
    call data%allocate_sf_int (sf_int)
    call sf_int%init (data)
    call sf_int%set_beam_index ([1])
    
    call sf_int%write (u)

    write (u, "(A)")
    write (u, "(A)")  "* Initialize incoming momentum with E=500"
    write (u, "(A)")
    E = 500
    k = vector4_moving (E, sqrt (E**2 - flavor_get_mass (flv)**2), 3)
    call vector4_write (k, u)
    call sf_int%seed_kinematics ([k])

    write (u, "(A)")
    write (u, "(A)")  "* Set kinematics for x=0.5"
    write (u, "(A)")

    allocate (r (data%get_n_par ()))
    allocate (x (size (r)))

    r = 0.5_default
    call sf_int%complete_kinematics (x, f, r, map=.false.)
    call sf_int%write (u)

    write (u, "(A)")
    write (u, "(A,9(1x,F12.9))")  "x =", x
    write (u, "(A,9(1x,F12.9))")  "f =", f

    write (u, "(A)")
    write (u, "(A)")  "* Recover x from momenta"
    write (u, "(A)")

    q = interaction_get_momenta (sf_int%interaction_t, outgoing=.true.)
    call sf_int%final ()
    deallocate (sf_int)

    call data%allocate_sf_int (sf_int)
    call sf_int%init (data)
    call sf_int%set_beam_index ([1])

    call sf_int%seed_kinematics ([k])
    call interaction_set_momenta (sf_int%interaction_t, q, outgoing=.true.)
    call sf_int%recover_x (x)

    write (u, "(A,9(1x,F12.9))")  "x =", x

    write (u, "(A)")
    write (u, "(A)")  "* Evaluate for Q = 100 GeV"
    write (u, "(A)")

    call sf_int%complete_kinematics (x, f, r, map=.false.) 
    call sf_int%apply (scale = 100._default)
    call sf_int%write (u)


    write (u, "(A)")
    write (u, "(A)")  "* Cleanup"

    call sf_int%final ()
    call model_list_final ()
    
    write (u, "(A)")
    write (u, "(A)")  "* Test output end: sf_lhapdf_2"

  end subroutine sf_lhapdf_2


end module sf_lhapdf
